function P = OLHash(x1,x2,y,P,para)
% x1 and x2 is the data pair (column vec)
%   if kernel trick is applied. x1 and x2 should be kernelized by the
%   kernel trick
% y is the similarity label: 1 = similar; -1 = dissimilar
% P is the current model
%   P(i).W is the projection matrix; i<= para.T. 
% para is the parameter, including alpha, beta, c,r,T
%   the meaning of alpha, beta, c, r, T, please refer to the paper
%   if T = 1, it is a single model Online Hashing; when T> 1, it is a multi
%       models Online Hashing


% binary code prediction for old models
for i = 1:para.T
    code(i).h = 2* double(P(i).W'*[x1 x2] >= 0) -1;
end

% obtain risk(a.k.a. similarity loss in paper)
if y == 1
    for i = 1:para.T
        Dh = ( para.r - code(i).h(:,1)'*code(i).h(:,2) )/2; %Hamming dist
        risk_list(i) = max( (Dh - para.alpha) ,0); % risk
    end
    [risk idx] = min(risk_list); 
else
    for i = 1:para.T
        Dh = ( para.r - code(i).h(:,1)'*code(i).h(:,2) )/2; %Hamming dist
        risk_list(i) = max((para.beta* para.r - Dh) ,0); % risk
    end
    risk = max(risk_list); 
end


% update the model if risk is positive
if risk ~= 0
    if y == 1
        % sim case, only one model is required to be zero-risk
        
        h = code(idx).h;
        W = P(idx).W;
        tmpH = h .* (W'*[x1 x2]); % potential loss (Eq 19)
        
        % remove K0, reserve K1
        K0 = find(h(:,1) == h(:,2));
        tmpH(K0,:) = inf; 
        
        [delta f_1or2] = min(tmpH,[],2);
        
        [tmpV tmpI] = sort(delta,'ascend');
        
        % obtain zero-loss code
        g = h;
        for i = 1:ceil(risk)
            g(tmpI(i),f_1or2(tmpI(i))) = - h(tmpI(i),f_1or2(tmpI(i)));
        end
        
        % updat model
        loss = (h(:,1)-g(:,1))'*W'*x1 + (h(:,2)-g(:,2))'*W'*x2;
        dir = [x1 x2]*(g -h)';       
        tao = min(para.c,loss/norm(dir,'fro')^2);
        
        P(idx).W = W + tao * dir;
        
        %clear h W tmpH K1 delta f_1or2 tmpV tmpI g loss dir tao
        
        
    else % y = -1
        % dissim case, all models are required to be zero-risk
        
        idx = find(risk_list > 0);
        
        for j = 1:length(idx)
            h = code(idx(j)).h;
            W = P(idx(j)).W;
            tmpH = h .* (W'*[x1 x2]); % potential loss (Eq 19)
            
            % remove K1, reserve K0
            K1 = find(h(:,1) ~= h(:,2));
            tmpH(K1,:) = inf; 

            [delta f_1or2] = min(tmpH,[],2);

            [tmpV tmpI] = sort(delta,'ascend');

            % obtain zero-loss code
            g = h;
            for i = 1:ceil(risk_list(j))
                g(tmpI(i),f_1or2(tmpI(i))) = - h(tmpI(i),f_1or2(tmpI(i)));
            end
            % updat model
            loss = (h(:,1)-g(:,1))'*W'*x1 + (h(:,2)-g(:,2))'*W'*x2;
            dir = [x1 x2]*(g -h)';       
            tao = min(para.c,loss/norm(dir,'fro')^2);

            P(idx(j)).W = W + tao * dir;
            
            clear h W tmpH K1 delta f_1or2 tmpV tmpI g loss dir tao
        end
        
    end
    
end